/* This file implements kanji 'radical' search widgets.
 * At one point, I intended to make my own radical list, with
 * my own definition of radicals
 * But I have wimped out and decided to go with the "standard" ones.
 *
 * This stuff takes advantage of the 'radkfile' file, included with
 * Jim Breen's xjdic program, and copied here with permission.
 * 
 */


#include <stdio.h>

#include <Xos.h>
#include <Xfuncs.h>
#include <Intrinsic.h>
#include <StringDefs.h>

#include <Shell.h>

#include <Xaw/Command.h>
#include <Xaw/Label.h>
#include <Xaw/Form.h>
#include <Xaw/Box.h>
#include <Xaw/AsciiText.h>

#include "defs.h"
#include "externs.h"
#include "game.h"    /* for GUESS_XXX defines */
#include "search.h"
#include "multikanji.h"
#include "searchwidgets.h"
#include "utils.h"
#include "init.h" /* for setup_deletewindow() */


Widget radical_popup;

Widget radsearch_help; /* like a status bar */

int num_radicals=0; /* number of radicals read from file */


#define MAXRADICALS 250
#define MAXKRADPOST 3200

XChar2b radicals[MAXRADICALS]; /* char for each radical */
XChar2b radical_array[MAXRADICALS][MAXKRADPOST];
Widget radbuttons[MAXRADICALS];
CARD8 radtoggle[MAXRADICALS];   /* bitmap to show if radical is selected*/

int numradsactive; /* number of radicals indicated by current select */
XChar2b activelist[MAXRADICALS];


static void setradstatus(char *s)
{
	XtVaSetValues(radsearch_help,XtNlabel,s,NULL);
}


/* Callback for 'clear' button
 * Note that we have to duplicate PART of functionality in here,
 * in subtractRadical()
 */
void ClearRadicals(Widget w, XtPointer data, XtPointer call_data)
{
	int rcount;
	for(rcount=0; rcount<num_radicals; rcount++){
		UnhighlightButton(radbuttons[rcount]);
		radtoggle[rcount]=0;
	}
	numradsactive=0;
	activelist[0].byte1=0;
	setradstatus("No radicals selected");
	ClearAllMulti();
	
}


/* given a radical indexnumber, 
 * take all kanji referenced by that radical and add to 'active' list.
 * return 1 if okay, 0 if we tried to merge, and failed.
 */
static int mergeRadical(int radnum){
	XChar2b newlist[MAXRADICALS];
	XChar2b searchstr[2];
	XChar2b *kradptr;
	int newcount;
	newcount=0;
	searchstr[1].byte1=0;

	if(radnum>=num_radicals){
		puts("ERROR: mergeRadical passed invalid radnum");
		return 0;
	}

	kradptr=radical_array[radnum];

	while(kradptr->byte1!=0){
		searchstr[0]=*kradptr;
		if(matchkanji(activelist,searchstr)){
			newlist[newcount++]=*kradptr;
		}
		kradptr++;
	}
	if(newcount>0){
		newlist[newcount].byte1=0;
		bcopy(newlist,activelist,(newcount+1)*sizeof(XChar2b));
		return 1;
	} else {
		return 0;
	}
	
}

/* given a radical indexnumber, 
 * take all kanji referenced by that radical and add to 'active' list.
 * return 1 if okay, 0 if we tried to merge, and failed.
 */
static int addRadical(int radnum)
{
	int radcount;
	XChar2b *radptr;
	if((radnum>num_radicals)||(radnum<0)){
		puts("ERROR: addRadical passed invalid number");
		return 0;
	}

	if(numradsactive==0){
		bcopy(radical_array[radnum], activelist,
		       MAXRADICALS*sizeof(XChar2b));
	} else {
		if(!mergeRadical(radnum)){
			setradstatus("no combination possible");
			return 0;
		}
	}



	/* Now just tally up active list, and we're done */
	radptr=activelist; radcount=0;
	while(radptr->byte1!=0){
		radcount++;
		radptr++;
	}
	
	{
		char countbuf[50];
		sprintf(countbuf,"%d kanji possible",radcount);
		setradstatus(countbuf);
		numradsactive=radcount;
	}

	return 1;
}


/* Need to just recalculate intersection of all currently active
 * radicals, except the one passed in.
 * Make sure to update numradsactive
 * Logic in here must match ClearRadicals()
 */
static void subtractRadical(int radnum)
{
	int rcount,totalactive;

	radtoggle[radnum]=0;
	numradsactive=0;
	activelist[0].byte1=0;

	/* Kinda brute-force, but cant think of "elegant" way to do this */

	for(rcount=0,totalactive=0; rcount<num_radicals; rcount++){
		if(radtoggle[rcount]){
			addRadical(rcount);
			totalactive++;
		}
	}

	/* Thats all there is to it. addRadical() will have
	 * updated numradsactive for us
	 */
	if(totalactive==0){
		setradstatus("No radicals selected");
	}

}


/* callback for when user clicks on an individual radical button */
void SelectRadical(Widget w, XtPointer data, XtPointer call_data)
{
	int rcount=0;
	while(rcount<num_radicals){
		if(w==radbuttons[rcount]){
			break;
		}
		rcount++;
	}
	if(rcount==num_radicals){
		puts("Error: unknown radicalbutton called SelectRadical?");
		return;
	}
#ifdef DEBUG
	printf("SelectRadical called by radical %d\n",rcount);
#endif

	if(radtoggle[rcount]){
		UnhighlightButton(radbuttons[rcount]);
		subtractRadical(rcount);
	} else {
		if(!addRadical(rcount)){
			return;
		}
		HighlightButton(radbuttons[rcount]);
		radtoggle[rcount]=1;
	}


	if(numradsactive<100){
		TRANSLATION kanjiptr;
		ClearAllMulti();

		for(rcount=0;rcount<numradsactive;rcount++){
			int kindex=(activelist[rcount].byte1 <<8) +
			            activelist[rcount].byte2;
			kanjiptr=translations[kindex];
			AddMultiTranslation(kanjiptr);
		}
		SetMultiMode(GUESS_KANJI);
		ShowMulti();
	}
}


XChar2b fakelabel[2]={{0x24, 0x22}, {0x0, 0x0}};

/*
 * Internal routine, to create the buttons for the
 * kanji radical search popup.
 * Should have:
 *  -  Buttons for all 250 radicals
 *  -  a 'clear' button
 *  -  a 'search' button
 *  -  a status label.
 *
 */
void makeradicalinput(Widget parent)
{
	Widget radform,bottomform;
	Widget clearbutton;
	int radcount;
	XChar2b buttonlabel[2];
	buttonlabel[1].byte1=0;
	buttonlabel[1].byte2=0;

	radform=XtVaCreateManagedWidget("radinputform", boxWidgetClass,parent,
		XtNwidth,600,
		NULL);

	for(radcount=0;radcount<num_radicals;radcount++){
		char radbname[10];
		sprintf(radbname,"radb%d\n",radcount);
		buttonlabel[0].byte1=radicals[radcount].byte1;
		buttonlabel[0].byte2=radicals[radcount].byte2;
		radbuttons[radcount]=XtVaCreateManagedWidget(radbname,
		        commandWidgetClass, radform,
			XtNlabel,buttonlabel,
			XtNencoding, XawTextEncodingChar2b,
			XtNfont, smallkfont,
		        NULL);

		XtAddCallback(radbuttons[radcount], XtNcallback,
				      SelectRadical, NULL);
	}

	bottomform=XtVaCreateManagedWidget("radcommandform", formWidgetClass,
		parent,
		XtNfromVert,radform,
		XtNborderWidth,0,
		NULL);

	clearbutton=XtVaCreateManagedWidget("radclear",
	        commandWidgetClass, bottomform,
		XtNlabel,"clear",
	        NULL);

	radsearch_help=XtVaCreateManagedWidget("radsearchhelp",
		labelWidgetClass, bottomform,
		XtNlabel,"                  Radical Search                    ",
		XtNfromHoriz,clearbutton,
	        NULL);
	
	XtAddCallback(clearbutton, XtNcallback,
				      ClearRadicals, NULL);

}

/* exported routine */
/* Lets other functions know if radical search is available or not.
 */ 
int HaveRadicals()
{
	if(num_radicals>0) return 1;
	return 0;
}

/* exported routine */
/* Call HaveRadicals() if you want to know if it
 * succeeded or not
 */
void InitRadicals()
{
	/* I'm not particularly proud of the internals of this thing.
	 * Unfortunately, the file format doesnt easily lend itself
	 * to a better approach, that I can think of.
	 * Oh well, memory is a lot cheaper these days. Sigh.
	 */
	char *radfile=RADLOCATION; /* location is defined in Imakefile */
	char linebuf[1024];
	int cur_radnum=-1;
	int kradpos=0;
	FILE *radfp;

	numradsactive=0;
	bzero(radical_array,MAXRADICALS*MAXKRADPOST*sizeof(XChar2b));
	bzero(radtoggle,MAXRADICALS);

	radfp=fopen(radfile,"r");
	if(radfp==NULL){
		num_radicals=0;
		return;
	}

	while(cur_radnum<MAXRADICALS){
		int linepos=0;
		if(fgets(linebuf, 1024, radfp)==NULL){
			break;
		}
		if(linebuf[0]=='#'){
			continue;
		}
		if(linebuf[0]=='$'){
			cur_radnum++;
			kradpos=0;
			radicals[cur_radnum].byte1=linebuf[2]&0x7f;
			radicals[cur_radnum].byte2=linebuf[3]&0x7f;

			continue;
		}
		/* otherwise, must be radical list line. arg */
		linepos=0;
		while(linebuf[linepos]!='\n'){
			radical_array[cur_radnum][kradpos].byte1=
			      linebuf[linepos++]&0x7f;
			radical_array[cur_radnum][kradpos].byte2=
			      linebuf[linepos++]&0x7f;
			kradpos++;
		}
	}

	if(cur_radnum==MAXRADICALS){
		fprintf(stderr,"ERROR:InitRadicals exceeded MAXRADICALS\n");
		num_radicals=0;
		return;
	}
	num_radicals=cur_radnum;

	printf("%d radicals read from radkfile\n",num_radicals);

}


/* exported routine */
void MakeRadicalinputPopup()
{
	Widget radicalinputform;
	if(num_radicals==0) return;

	radical_popup = XtVaCreatePopupShell("kdrill_radicalsearch ",
		transientShellWidgetClass,
		search_popup,
		NULL);

	radicalinputform = XtVaCreateManagedWidget("radicalinputform",
					     formWidgetClass,
					     radical_popup,
					     NULL);

	/* make all the substuff */
	makeradicalinput(radicalinputform);

}

/*
 * Exported routine:
 * 
 * It pops up the kanji radical search window
 */
void ShowRadicalinput(Widget w,XtPointer client_data, XtPointer call_data)
{
	static int radicals_up = -1;

	static Position rel_x,rel_y;
	Position x,y;

	if(num_radicals==0){
		setstatus("ERROR: Radical file radkfile not present");
		return;
	}

	if(radicals_up==-1){
		/* first time init.. */
		/*rel_x = GetXtNumber("radical_popupx","Search_popupx");*/
		/*rel_y = GetXtNumber("radical_popupy","Search_popupy");*/
		rel_x = 10;
		rel_y = 10;
		radicals_up=0;
	}
	if(isMapped(radical_popup)==False){
		XtTranslateCoords(search_popup,rel_x,rel_y,&x,&y);
		XtVaSetValues(radical_popup,
		      XtNx,x,
		      XtNy,y,
		      NULL);
		XtPopup(radical_popup,XtGrabNone);
		if(radicals_up==0){
			setup_deletewindow(radical_popup);
			radicals_up=1;
		}
		
		setstatus("Bringing up radical input window...");
	} else {
		XtPopdown(radical_popup);
	}
}


